@@ -30,11 +30,6 @@ filtering network based on user submissions of spam.  Detection is done
 with signatures that efficiently spot mutating spam content and user
 input is validated through reputation assignments.
 
-Note that Razor2 is disabled by default in C<init.pre> because it is not
-available for unlimited free use.  It is currently free for personal
-use, subject to capacity constraints.  See the Cloudmark SpamNet Service
-Policy for more details.
-
 See http://razor.sourceforge.net/ for more information about Razor.
 
 =head1 USER SETTINGS
@@ -48,12 +43,17 @@ package Mail::SpamAssassin::Plugin::Razor2;
 use Mail::SpamAssassin::Plugin;
 use Mail::SpamAssassin::Logger;
 use Mail::SpamAssassin::Timeout;
+use Mail::SpamAssassin::SubProcBackChannel;
+use Mail::SpamAssassin::Util qw(force_die am_running_on_windows);
 use strict;
 use warnings;
-use bytes;
+# use bytes;
+use re 'taint';
 
-use vars qw(@ISA);
-@ISA = qw(Mail::SpamAssassin::Plugin);
+use Storable;
+use POSIX qw(PIPE_BUF WNOHANG);
+
+our @ISA = qw(Mail::SpamAssassin::Plugin);
 
 sub new {
   my $class = shift;
@@ -78,8 +78,8 @@ sub new {
     }
   }
 
-  $self->register_eval_rule("check_razor2");
-  $self->register_eval_rule("check_razor2_range");
+  $self->register_eval_rule("check_razor2", $Mail::SpamAssassin::Conf::TYPE_FULL_EVALS);
+  $self->register_eval_rule("check_razor2_range", $Mail::SpamAssassin::Conf::TYPE_FULL_EVALS);
 
   $self->set_config($mailsaobject->{conf});
 
@@ -88,7 +88,7 @@ sub new {
 
 sub set_config {
   my ($self, $conf) = @_;
-  my @cmds = ();
+  my @cmds;
 
 =item use_razor2 (0|1)		(default: 1)
 
@@ -98,10 +98,32 @@ Whether to use Razor2, if it is available.
 
   push(@cmds, {
     setting => 'use_razor2',
+    is_admin => 1,
     default => 1,
     type => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
   });
 
+=item razor_fork (0|1)		(default: 1)
+
+Instead of running Razor2 synchronously, fork separate process for it and
+read the results in later (similar to async DNS lookups). Increases
+throughput. Considered experimental on Windows, where default is 0.
+
+=cut
+
+  push(@cmds, {
+    setting => 'razor_fork',
+    is_admin => 1,
+    default => am_running_on_windows()?0:1,
+    type => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
+  });
+
+=back
+
+=head1 ADMINISTRATOR SETTINGS
+
+=over 4
+
 =item razor_timeout n		(default: 5)
 
 How many seconds you wait for Razor to complete before you go on without
@@ -111,8 +133,9 @@ the results
 
   push(@cmds, {
     setting => 'razor_timeout',
+    is_admin => 1,
     default => 5,
-    type => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
+    type => $Mail::SpamAssassin::Conf::CONF_TYPE_DURATION,
   });
 
 =item razor_config filename
@@ -124,6 +147,7 @@ Currently this is left to Razor to decide.
 
   push(@cmds, {
     setting => 'razor_config',
+    is_admin => 1,
     type => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING,
   });
 
@@ -131,10 +155,10 @@ Currently this is left to Razor to decide.
 }
 
 sub razor2_access {
-  my ($self, $fulltext, $type) = @_;
+  my ($self, $fulltext, $type, $deadline) = @_;
   my $timeout = $self->{main}->{conf}->{razor_timeout};
   my $return = 0;
-  my @results = ();
+  my @results;
 
   my $debug = $type eq 'check' ? 'razor2' : 'reporter';
 
@@ -146,7 +170,10 @@ sub razor2_access {
 
   Mail::SpamAssassin::PerMsgStatus::enter_helper_run_mode($self);
 
-  my $timer = Mail::SpamAssassin::Timeout->new({ secs => $timeout });
+  my $rnd = rand(0x7fffffff);  # save entropy before Razor clobbers it
+
+  my $timer = Mail::SpamAssassin::Timeout->new(
+               { secs => $timeout, deadline => $deadline });
   my $err = $timer->run_and_catch(sub {
 
     local ($^W) = 0;    # argh, warnings in Razor
@@ -189,7 +216,7 @@ sub razor2_access {
       my $sigs = $rc->compute_sigs($objects)
 	  or die "$debug: error in compute_sigs";
 
-      # if mail isn't whitelisted, check it out
+      # if mail isn't welcomelisted, check it out
       # see 'man razor-whitelist'
       if ($type ne 'check' || ! $rc->local_check($objects->[0])) {
 	# provide a better error message when servers are unavailable,
@@ -236,7 +263,9 @@ sub razor2_access {
             last;
           }
         }
-        close $rc->{logref}->{fd} if ($untie);
+        if ($untie) {
+          close($rc->{logref}->{fd})  or die "error closing log: $!";
+        }
       }
 
       if ($type eq 'check') {
@@ -297,7 +326,10 @@ sub razor2_access {
   # OK, that's enough Razor stuff. now, reset all that global
   # state it futzes with :(
   # work around serious brain damage in Razor2 (constant seed)
-  srand;
+  $rnd ^= int(rand(0xffffffff));  # mix old acc with whatever came out of razor
+  srand;                          # let Perl give it a try ...
+  $rnd ^= int(rand(0xffffffff));  # ... and mix-in that too
+  srand($rnd & 0x7fffffff);  # reseed, keep it unsigned 32-bit just in case
 
   Mail::SpamAssassin::PerMsgStatus::leave_helper_run_mode($self);
 
@@ -336,7 +368,9 @@ sub plugin_report {
   return unless $self->{main}->{conf}->{use_razor2};
   return if $options->{report}->{options}->{dont_report_to_razor};
 
-  if ($self->razor2_access($options->{text}, 'report')) {
+  my $timer = $self->{main}->time_method("razor2_report");
+
+  if ($self->razor2_access($options->{text}, 'report', undef)) {
     $options->{report}->{report_available} = 1;
     info('reporter: spam reported to Razor');
     $options->{report}->{report_return} = 1;
@@ -349,46 +383,208 @@ sub plugin_report {
 sub plugin_revoke {
   my ($self, $options) = @_;
 
+  my $timer = $self->{main}->time_method("razor2_revoke");
+
   return unless $self->{razor2_available};
   return if $self->{main}->{local_tests_only};
   return unless $self->{main}->{conf}->{use_razor2};
   return if $options->{revoke}->{options}->{dont_report_to_razor};
 
-  if ($self->razor2_access($options->{text}, 'revoke')) {
+  if ($self->razor2_access($options->{text}, 'revoke', undef)) {
     $options->{revoke}->{revoke_available} = 1;
-    dbg('reporter: spam revoked from Razor');
+    info('reporter: spam revoked from Razor');
     $options->{revoke}->{revoke_return} = 1;
   }
   else {
-    dbg('reporter: could not revoke spam from Razor');
+    info('reporter: could not revoke spam from Razor');
+  }
+}
+
+sub finish_parsing_start {
+  my ($self, $opts) = @_;
+
+  # If forking, hard adjust priority -100 to launch early
+  # Find rulenames from eval_to_rule mappings
+  if ($opts->{conf}->{razor_fork}) {
+    foreach (@{$opts->{conf}->{eval_to_rule}->{check_razor2}}) {
+      dbg("razor2: adjusting rule $_ priority to -100");
+      $opts->{conf}->{priority}->{$_} = -100;
+    }
+    foreach (@{$opts->{conf}->{eval_to_rule}->{check_razor2_range}}) {
+      dbg("razor2: adjusting rule $_ priority to -100");
+      $opts->{conf}->{priority}->{$_} = -100;
+    }
   }
 }
 
 sub check_razor2 {
-  my ($self, $permsgstatus, $full) = @_;
+  my ($self, $pms, $full) = @_;
 
-  return $permsgstatus->{razor2_result} if (defined $permsgstatus->{razor2_result});
-  $permsgstatus->{razor2_result} = 0;
-  $permsgstatus->{razor2_cf_score} = { '4' => 0, '8' => 0 };
+  return 0 unless $self->{razor2_available};
+  return 0 unless $self->{main}->{conf}->{use_razor2};
 
-  return unless $self->{razor2_available};
-  return unless $self->{main}->{conf}->{use_razor2};
+  return $pms->{razor2_result} if (defined $pms->{razor2_result});
 
-  my $return;
-  my @results = ();
+  return 0 if $pms->{razor2_running};
+  $pms->{razor2_running} = 1;
 
-  # TODO: check for cache header, set results appropriately
+  my $timer = $self->{main}->time_method("check_razor2");
+
+  ## non-forking method
+
+  if (!$self->{main}->{conf}->{razor_fork}) {
+    # TODO: check for cache header, set results appropriately
+    # do it this way to make it easier to get out the results later from the
+    # netcache plugin ... what netcache plugin?
+    (undef, my @results) =
+      $self->razor2_access($full, 'check', $pms->{master_deadline});
+    return $self->_check_result($pms, \@results);
+  }
+
+  ## forking method
+
+  $pms->{razor2_rulename} = $pms->get_current_eval_rule_name();
+
+  # create socketpair for communication
+  $pms->{razor2_backchannel} = Mail::SpamAssassin::SubProcBackChannel->new();
+  my $back_selector = '';
+  $pms->{razor2_backchannel}->set_selector(\$back_selector);
+  eval {
+    $pms->{razor2_backchannel}->setup_backchannel_parent_pre_fork();
+  } or do {
+    dbg("razor2: backchannel pre-setup failed: $@");
+    delete $pms->{razor2_backchannel};
+    return 0;
+  };
+
+  my $pid = fork();
+  if (!defined $pid) {
+    info("razor2: child fork failed: $!");
+    delete $pms->{razor2_backchannel};
+    return 0;
+  }
+  if (!$pid) {
+    $0 = "$0 (razor2)";
+    $SIG{CHLD} = 'DEFAULT';
+    $SIG{PIPE} = 'IGNORE';
+    $SIG{$_} = sub {
+      eval { dbg("razor2: child process $$ caught signal $_[0]"); };
+      force_die(6);  # avoid END and destructor processing
+      } foreach am_running_on_windows()?qw(INT HUP TERM QUIT):qw(INT HUP TERM TSTP QUIT USR1 USR2);
+    dbg("razor2: child process $$ forked");
+    $pms->{razor2_backchannel}->setup_backchannel_child_post_fork();
+    (undef, my @results) =
+      $self->razor2_access($full, 'check', $pms->{master_deadline});
+    my $backmsg;
+    eval {
+      $backmsg = Storable::freeze(\@results);
+    };
+    if ($@) {
+      dbg("razor2: child return value freeze failed: $@");
+      force_die(0); # avoid END and destructor processing
+    }
+    if (!syswrite($pms->{razor2_backchannel}->{parent}, $backmsg)) {
+      dbg("razor2: child backchannel write failed: $!");
+    }
+    force_die(0); # avoid END and destructor processing
+  }
+
+  $pms->{razor2_pid} = $pid;
+
+  eval {
+    $pms->{razor2_backchannel}->setup_backchannel_parent_post_fork($pid);
+  } or do {
+    dbg("razor2: backchannel post-setup failed: $@");
+    delete $pms->{razor2_backchannel};
+    return 0;
+  };
+
+  return; # return undef for async status
+}
+
+sub check_tick {
+  my ($self, $opts) = @_;
+  $self->_check_forked_result($opts->{permsgstatus}, 0);
+}
+
+sub check_cleanup {
+  my ($self, $opts) = @_;
+  $self->_check_forked_result($opts->{permsgstatus}, 1);
+}
+
+sub _check_forked_result {
+  my ($self, $pms, $finish) = @_;
+
+  return 0 if !$pms->{razor2_backchannel};
+  return 0 if !$pms->{razor2_pid};
+
+  my $timer = $self->{main}->time_method("check_razor2");
+
+  $pms->{razor2_abort} = $pms->{deadline_exceeded} || $pms->{shortcircuited};
+
+  my $kid_pid = $pms->{razor2_pid};
+  # if $finish, force waiting for the child
+  my $pid = waitpid($kid_pid, $finish && !$pms->{razor2_abort} ? 0 : WNOHANG);
+  if ($pid == 0) {
+    #dbg("razor2: child process $kid_pid not finished yet, trying later");
+    if ($pms->{razor2_abort}) {
+      dbg("razor2: bailing out due to deadline/shortcircuit");
+      kill('TERM', $kid_pid);
+      if (waitpid($kid_pid, WNOHANG) == 0) {
+        sleep(1);
+        if (waitpid($kid_pid, WNOHANG) == 0) {
+          dbg("razor2: child process $kid_pid still alive, KILL");
+          kill('KILL', $kid_pid);
+          waitpid($kid_pid, 0);
+        }
+      }
+      delete $pms->{razor2_pid};
+      delete $pms->{razor2_backchannel};
+    }
+    return 0;
+  } elsif ($pid == -1) {
+    # child does not exist?
+    dbg("razor2: child process $kid_pid already handled?");
+    delete $pms->{razor2_backchannel};
+    return 0;
+  }
+
+  $pms->rule_ready($pms->{razor2_rulename}); # mark rule ready for metas
+
+  dbg("razor2: child process $kid_pid finished, reading results");
+
+  my $backmsg;
+  my $ret = sysread($pms->{razor2_backchannel}->{latest_kid_fh}, $backmsg, am_running_on_windows()?512:PIPE_BUF);
+  if (!defined $ret || $ret == 0) {
+    dbg("razor2: could not read result from child: ".($ret == 0 ? 0 : $!));
+    delete $pms->{razor2_backchannel};
+    return 0;
+  }
+
+  delete $pms->{razor2_backchannel};
+
+  my $results;
+  eval {
+    $results = Storable::thaw($backmsg);
+  };
+  if ($@) {
+    dbg("razor2: child return value thaw failed: $@");
+    return;
+  }
+
+  $self->_check_result($pms, $results);
+}
+
+sub _check_result {
+  my ($self, $pms, $results) = @_;
 
-  # do it this way to make it easier to get out the results later from the
-  # netcache plugin
-  ($return, @results) = $self->razor2_access($full, 'check');
   $self->{main}->call_plugins ('process_razor_result',
-  	{ results => \@results, permsgstatus => $permsgstatus }
+  	{ results => $results, permsgstatus => $pms }
   );
 
-  foreach my $result (@results) {
+  foreach my $result (@$results) {
     if (exists $result->{result}) {
-      $permsgstatus->{razor2_result} = $result->{result} if $result->{result};
+      $pms->{razor2_result} = $result->{result} if $result->{result};
     }
     elsif ($result->{noresponse}) {
       dbg('razor2: part=' . $result->{part} . ' noresponse');
@@ -401,46 +597,86 @@ sub check_razor2 {
 
       next if $result->{contested};
 
-      my $cf = $permsgstatus->{razor2_cf_score}->{$result->{engine}} || 0;
+      my $cf = $pms->{razor2_cf_score}->{$result->{engine}} || 0;
       if ($result->{confidence} > $cf) {
-        $permsgstatus->{razor2_cf_score}->{$result->{engine}} = $result->{confidence};
+        $pms->{razor2_cf_score}->{$result->{engine}} = $result->{confidence};
       }
     }
   }
 
-  dbg("razor2: results: spam? " . $permsgstatus->{razor2_result});
-  while(my ($engine, $cf) = each %{$permsgstatus->{razor2_cf_score}}) {
+  $pms->{razor2_result} ||= 0;
+  $pms->{razor2_cf_score} ||= {};
+
+  dbg("razor2: results: spam? " . $pms->{razor2_result});
+  while(my ($engine, $cf) = each %{$pms->{razor2_cf_score}}) {
     dbg("razor2: results: engine $engine, highest cf score: $cf");
   }
 
-  return $permsgstatus->{razor2_result};
+  if ($self->{main}->{conf}->{razor_fork}) {
+    # forked needs to run got_hit()
+    if ($pms->{razor2_rulename} && $pms->{razor2_result}) {
+      $pms->got_hit($pms->{razor2_rulename}, "", ruletype => 'eval');
+    }
+    # forked needs to run range callbacks
+    if ($pms->{razor2_range_callbacks}) {
+      foreach (@{$pms->{razor2_range_callbacks}}) {
+        $self->check_razor2_range($pms, '', @$_);
+      }
+    }
+  }
+
+  return $pms->{razor2_result};
 }
 
 # Check the cf value of a given message and return if it's within the
 # given range
 sub check_razor2_range {
-  my ($self, $permsgstatus, $body, $engine, $min, $max) = @_;
+  my ($self, $pms, $body, $engine, $min, $max, $rulename) = @_;
 
   # If Razor2 isn't available, or the general test is disabled, don't
   # continue.
-  return unless $self->{razor2_available};
-  return unless $self->{main}->{conf}->{use_razor2};
-  return unless $self->{main}->{conf}->{scores}->{'RAZOR2_CHECK'};
+  return 0 unless $self->{razor2_available};
+  return 0 unless $self->{main}->{conf}->{use_razor2};
 
-  # If Razor2 hasn't been checked yet, go ahead and run it.
-  unless (defined $permsgstatus->{razor2_result}) {
-    $self->check_razor2($permsgstatus, $body);
+  # Check if callback overriding rulename
+  if (!defined $rulename) {
+    $rulename = $pms->get_current_eval_rule_name();
   }
 
+  if ($pms->{razor2_abort}) {
+    $pms->rule_ready($rulename); # mark rule ready for metas
+    return;
+  }
+
+  # If forked, call back later unless results are in
+  if ($self->{main}->{conf}->{razor_fork}) {
+    if (!defined $pms->{razor2_result}) {
+      dbg("razor2: delaying check_razor2_range call for $rulename");
+      # array matches check_razor2_range() argument order
+      push @{$pms->{razor2_range_callbacks}},
+        [$engine, $min, $max, $rulename];
+      return; # return undef for async status
+    }
+  } else {
+    # If Razor2 hasn't been checked yet, go ahead and run it.
+    # (only if we are non-forking.. forking will handle these in
+    # callbacks)
+    if (!$pms->{razor2_running}) {
+      $self->check_razor2($pms, $body);
+    }
+  }
+
+  $pms->rule_ready($rulename); # mark rule ready for metas
+
   my $cf = 0;
   if ($engine) {
-    $cf = $permsgstatus->{razor2_cf_score}->{$engine};
-    return unless defined $cf;
+    $cf = $pms->{razor2_cf_score}->{$engine};
+    return 0 unless defined $cf;
   }
   else {
     # If no specific engine was given to the rule, find the highest cf
     # determined and use that
-    while(my ($engine, $ecf) = each %{$permsgstatus->{razor2_cf_score}}) {
+    while(my ($engine, $ecf) = each %{$pms->{razor2_cf_score}}) {
       if ($ecf > $cf) {
         $cf = $ecf;
       }
@@ -448,13 +684,20 @@ sub check_razor2_range {
   }
 
   if ($cf >= $min && $cf <= $max) {
-    $permsgstatus->test_log(sprintf("cf: %3d", $cf));
+    my $cf_str = sprintf("cf: %3d", $cf);
+    $pms->test_log($cf_str, $rulename);
+    if ($self->{main}->{conf}->{razor_fork}) {
+      $pms->got_hit($rulename, "", ruletype => 'eval');
+    }
     return 1;
   }
 
-  return;
+  return 0;
 }
 
+# Version features
+sub has_fork { 1 }
+
 1;
 
 =back
