@@ -17,7 +17,7 @@
 
 =head1 NAME
 
-Mail::SpamAssassin::SQLBasedAddrList - SpamAssassin SQL Based Auto Whitelist
+Mail::SpamAssassin::SQLBasedAddrList - SpamAssassin SQL Based Auto Welcomelist
 
 =head1 SYNOPSIS
 
@@ -38,16 +38,17 @@ A SQL based persistent address list implementation.
 See C<Mail::SpamAssassin::PersistentAddrList> for more information.
 
 Uses DBI::DBD module access to your favorite database (tested with
-MySQL, SQLite and PostgreSQL) to store user auto-whitelists.
+MySQL, SQLite and PostgreSQL) to store user auto-welcomelists.
 
 The default table structure looks like this:
 CREATE TABLE awl (
   username varchar(100) NOT NULL default '',
   email varchar(255) NOT NULL default '',
   ip varchar(40) NOT NULL default '',
-  count int(11) NOT NULL default '0',
+  msgcount int(11) NOT NULL default '0',
   totscore float NOT NULL default '0',
   signedby varchar(255) NOT NULL default '',
+  last_hit timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP,
   PRIMARY KEY (username,email,signedby,ip)
 ) TYPE=MyISAM;
 
@@ -75,18 +76,16 @@ package Mail::SpamAssassin::SQLBasedAddrList;
 
 use strict;
 use warnings;
-use bytes;
+# use bytes;
 use re 'taint';
 
 # Do this silliness to stop RPM from finding DBI as required
-BEGIN { require DBI;  import DBI; }
+BEGIN { require DBI;  DBI->import; }
 
 use Mail::SpamAssassin::PersistentAddrList;
 use Mail::SpamAssassin::Logger;
 
-use vars qw(@ISA);
-
-@ISA = qw(Mail::SpamAssassin::PersistentAddrList);
+our @ISA = qw(Mail::SpamAssassin::PersistentAddrList);
 
 =head2 new
 
@@ -125,7 +124,7 @@ sub new_checker {
 
   if (!$main->{conf}->{user_awl_dsn} ||
       !$main->{conf}->{user_awl_sql_table}) {
-    dbg("auto-whitelist: sql-based invalid config");
+    dbg("auto-welcomelist: sql-based invalid config");
     return;
   }
 
@@ -136,12 +135,12 @@ sub new_checker {
   my $dbh = DBI->connect($dsn, $dbuser, $dbpass, {'PrintError' => 0});
 
   if(!$dbh) {
-    info("auto-whitelist: sql-based unable to connect to database (%s) : %s",
+    info("auto-welcomelist: sql-based unable to connect to database (%s) : %s",
          $dsn, DBI::errstr);
     return;
   }
 
-  dbg("auto-whitelist: sql-based connected to $dsn");
+  dbg("auto-welcomelist: sql-based connected to $dsn");
 
   $self = { 'main'      => $main,
             'dsn'       => $dsn,
@@ -163,9 +162,9 @@ sub new_checker {
     }
   }
   $self->{_with_awl_signer} =
-    $main->{conf}->{auto_whitelist_distinguish_signed};
+    $main->{conf}->{auto_welcomelist_distinguish_signed};
 
-  dbg("auto-whitelist: sql-based using username: ".$self->{_username});
+  dbg("auto-welcomelist: sql-based using username: ".$self->{_username});
 
   return bless ($self, $class);
 }
@@ -193,7 +192,7 @@ sub get_addr_entry {
 
   my $entry = { addr     => $addr,
                 exists_p => 0,
-                count    => 0,
+                msgcount    => 0,
                 totscore => 0,
                 signedby => $signedby,
               };
@@ -202,7 +201,7 @@ sub get_addr_entry {
 
   return $entry  unless $email ne '' && (defined $ip || defined $signedby);
 
-  my $sql = "SELECT count, totscore FROM $self->{tablename} " .
+  my $sql = "SELECT msgcount, totscore FROM $self->{tablename} " .
             "WHERE username = ? AND email = ?";
   my @args = ( $email );
   if (!$self->{_with_awl_signer}) {
@@ -220,13 +219,22 @@ sub get_addr_entry {
     }
     push(@args, @signedby);
   }
+  $sql .= " ORDER BY last_hit";
+
   my $sth = $self->{dbh}->prepare($sql);
+
+  unless (defined($sth)) {
+    info("auto-welcomelist: sql-based get_addr_entry %s: SQL prepare error: %s",
+         join('|',@args), $self->{dbh}->errstr);
+    return $entry;
+  }
+
   my $rc = $sth->execute($self->{_username}, @args);
 
   if (!$rc) { # there was an error, but try to go on
-    info("auto-whitelist: sql-based get_addr_entry %s: SQL error: %s",
+    info("auto-welcomelist: sql-based get_addr_entry %s: SQL error: %s",
          join('|',@args), $sth->errstr);
-    $entry->{count} = 0;
+    $entry->{msgcount} = 0;
     $entry->{totscore} = 0;
   }
   else {
@@ -235,21 +243,22 @@ sub get_addr_entry {
     # how to combine data if there are several entries (like signed by
     # an author domain and by a remailer)?  for now just take an average
     while ( defined($aryref = $sth->fetchrow_arrayref()) ) {
-      if (defined $entry->{count} && defined $aryref->[1]) {
-        $entry->{count} += $aryref->[0];
-        $entry->{totscore} += $aryref->[1];
+      if (defined $entry->{msgcount} && defined $aryref->[1]) {
+        $entry->{msgcount} = $aryref->[0];
+        $entry->{totscore} = $aryref->[1];
       }
       $entry->{exists_p} = 1;
       $cnt++;
     }
-    dbg("auto-whitelist: sql-based get_addr_entry: %s for %s",
+    dbg("auto-welcomelist: sql-based get_addr_entry: %s for %s",
         $cnt ? "found $cnt entries" : 'no entries found',
         join('|',@args) );
   }
   $sth->finish();
 
-  dbg("auto-whitelist: sql-based %s scores %s, count %s",
-      join('|',@args), $entry->{totscore}, $entry->{count});
+  # tests t/sql_based_w*.t look for this dbg line in this format
+  dbg("auto-welcomelist: sql-based %s scores %.1f, msgcount %s",
+      join('|',@args), $entry->{totscore}, $entry->{msgcount});
 
   return $entry;
 }
@@ -276,7 +285,7 @@ sub add_score {
   
   my ($email, $ip) = $self->_unpack_addr($entry->{addr});
 
-  $entry->{count} += 1;
+  $entry->{msgcount} += 1;
   $entry->{totscore} += $score;
   my $signedby = $entry->{signedby};
   
@@ -287,7 +296,7 @@ sub add_score {
 
   my $inserted = 0;
 
-  { my @fields = qw(username email ip count totscore);
+  { my @fields = qw(username email ip msgcount totscore);
     my @signedby;
     if ($self->{_with_awl_signer}) {
       push(@fields, 'signedby');
@@ -297,26 +306,48 @@ sub add_score {
     my @args = ($self->{_username}, $email, $ip, 1, $score);
     my $sql = sprintf("INSERT INTO %s (%s) VALUES (%s)", $self->{tablename},
                       join(',', @fields),  join(',', ('?') x @fields));
+    if ($self->{dsn} =~ /^DBI:(?:pg|SQLite)/i) {
+       $sql .= " ON CONFLICT (username, email, signedby, ip) DO UPDATE set msgcount = ?, totscore = totscore + ?";
+    } elsif ($self->{dsn} =~ /^DBI:(?:mysql|MariaDB)/i) {
+       $sql .= " ON DUPLICATE KEY UPDATE msgcount = ?, totscore = totscore + ?";
+    }
+
     my $sth = $self->{dbh}->prepare($sql);
 
+    unless (defined($sth)) {
+      info("auto-welcomelist: sql-based add_score/insert %s: SQL prepare error: %s",
+           join('|',@args), $self->{dbh}->errstr);
+      return $entry;
+    }
+
     if (!$self->{_with_awl_signer}) {
-      my $rc = $sth->execute(@args);
+      my $rc;
+      if ($self->{dsn} =~ /^DBI:(?:pg|SQLite|mysql|MariaDB)/i) {
+          $rc = $sth->execute(@args, $entry->{msgcount}, $score);
+      } else {
+          $rc = $sth->execute(@args);
+      }
       if (!$rc) {
-        dbg("auto-whitelist: sql-based add_score/insert %s: SQL error: %s",
+        dbg("auto-welcomelist: sql-based add_score/insert %s: SQL error: %s",
              join('|',@args), $sth->errstr);
       } else {
-        dbg("auto-whitelist: sql-based add_score/insert ".
+        dbg("auto-welcomelist: sql-based add_score/insert ".
             "score %s: %s", $score, join('|',@args));
         $inserted = 1; $entry->{exists_p} = 1;
       }
     } else {
       for my $s (@signedby) {
-        my $rc = $sth->execute(@args, $s);
+        my $rc;
+	if ($self->{dsn} =~ /^DBI:(?:pg|SQLite|mysql|MariaDB)/i) {
+          $rc = $sth->execute(@args, $s, $entry->{msgcount}, $score);
+	} else {
+	  $rc = $sth->execute(@args, $s);
+	}
         if (!$rc) {
-          dbg("auto-whitelist: sql-based add_score/insert %s: SQL error: %s",
+          dbg("auto-welcomelist: sql-based add_score/insert %s: SQL error: %s",
               join('|',@args,$s), $sth->errstr);
         } else {
-          dbg("auto-whitelist: sql-based add_score/insert ".
+          dbg("auto-welcomelist: sql-based add_score/insert ".
               "score %s: %s", $score, join('|',@args,$s));
           $inserted = 1; $entry->{exists_p} = 1;
         }
@@ -324,13 +355,13 @@ sub add_score {
     }
   }
 
-  if (!$inserted) {
+  if (!$inserted && $self->{dsn} !~ /^DBI:(?:pg|SQLite|mysql|MariaDB)/i) {
     # insert failed, assume primary key constraint, so try the update
 
     my $sql = "UPDATE $self->{tablename} ".
-              "SET count = count + 1, totscore = totscore + ? ".
+              "SET msgcount = ?, totscore = totscore + ? ".
               "WHERE username = ? AND email = ?";
-    my(@args) = ($score, $self->{_username}, $email);
+    my(@args) = ($entry->{msgcount}, $score, $self->{_username}, $email);
     if ($self->{_with_awl_signer}) {
       my @signedby = !defined $signedby ? () : split(' ', lc $signedby);
       if (!@signedby) {
@@ -346,19 +377,26 @@ sub add_score {
     push(@args, $ip);
 
     my $sth = $self->{dbh}->prepare($sql);
+
+    unless (defined($sth)) {
+      info("auto-welcomelist: sql-based add_score/update %s: SQL prepare error: %s",
+           join('|',@args), $self->{dbh}->errstr);
+      return $entry;
+    }
+
     my $rc = $sth->execute(@args);
-    
+
     if (!$rc) {
-      info("auto-whitelist: sql-based add_score/update %s: SQL error: %s",
+      info("auto-welcomelist: sql-based add_score/update %s: SQL error: %s",
            join('|',@args), $sth->errstr);
     } else {
-      dbg("auto-whitelist: sql-based add_score/update ".
-          "new count: %s, new totscore: %s for %s",
-          $entry->{count}, $entry->{totscore}, join('|',@args));
+      dbg("auto-welcomelist: sql-based add_score/update ".
+          "new msgcount: %s, new totscore: %s for %s",
+          $entry->{msgcount}, $entry->{totscore}, join('|',@args));
       $entry->{exists_p} = 1;
     }
   }
-  
+
   return $entry;
 }
 
@@ -386,12 +424,12 @@ sub remove_entry {
   # when $ip is equal to none then attempt to delete all entries
   # associated with address
   if ($ip eq 'none') {
-    dbg("auto-whitelist: sql-based remove_entry: removing all entries matching $email");
+    dbg("auto-welcomelist: sql-based remove_entry: removing all entries matching $email");
   }
   else {
     $sql .= " AND ip = ?";
     push(@args, $ip);
-    dbg("auto-whitelist: sql-based remove_entry: removing single entry matching ".$entry->{addr});
+    dbg("auto-welcomelist: sql-based remove_entry: removing single entry matching ".$entry->{addr});
   }
   # if a key 'signedby' exists in the $entry, be selective on its value too
   my $signedby = $entry->{signedby};
@@ -406,10 +444,17 @@ sub remove_entry {
   }
 
   my $sth = $self->{dbh}->prepare($sql);
+
+  unless (defined($sth)) {
+    info("auto-welcomelist: sql-based remove_entry %s: SQL prepare error: %s",
+         join('|',@args), $self->{dbh}->errstr);
+    return;
+  }
+
   my $rc = $sth->execute(@args);
 
   if (!$rc) {
-    info("auto-whitelist: sql-based remove_entry %s: SQL error: %s",
+    info("auto-welcomelist: sql-based remove_entry %s: SQL error: %s",
          join('|',@args), $sth->errstr);
   }
   else {
@@ -430,7 +475,7 @@ This method provides the necessary cleanup for the address list.
 
 sub finish {
   my ($self) = @_;
-  dbg("auto-whitelist: sql-based finish: disconnected from " . $self->{dsn});
+  dbg("auto-welcomelist: sql-based finish: disconnected from " . $self->{dsn});
   $self->{dbh}->disconnect();
 }
 
@@ -439,7 +484,7 @@ sub finish {
 private instance (String, String) _unpack_addr(string $addr)
 
 Description:
-This method splits an autowhitelist address into it's two components,
+This method splits an autowelcomelist address into it's two components,
 email and ip address.
 
 =cut
@@ -450,7 +495,7 @@ sub _unpack_addr {
   my ($email, $ip) = split(/\|ip=/, $addr);
 
   unless ($email && $ip) {
-    dbg("auto-whitelist: sql-based _unpack_addr: unable to decode $addr");
+    dbg("auto-welcomelist: sql-based _unpack_addr: unable to decode $addr");
   }
 
   return ($email, $ip);
