@@ -21,29 +21,26 @@ package Mail::SpamAssassin::SpamdForkScaling;
 
 use strict;
 use warnings;
-use bytes;
+# use bytes;
 use re 'taint';
 use Errno qw();
+use Scalar::Util qw(blessed);
 
 use Mail::SpamAssassin::Util qw(am_running_on_windows);
 use Mail::SpamAssassin::Logger;
 use Mail::SpamAssassin::Timeout;
 
-use vars qw {
-  @PFSTATE_VARS %EXPORT_TAGS @EXPORT_OK
-};
-
 use base qw( Exporter );
 
-@PFSTATE_VARS = qw(
+our @PFSTATE_VARS = qw(
   PFSTATE_ERROR PFSTATE_STARTING PFSTATE_IDLE PFSTATE_BUSY PFSTATE_KILLED
   PFORDER_ACCEPT PFSTATE_GOT_SIGCHLD
 );
 
-%EXPORT_TAGS = (
+our %EXPORT_TAGS = (
   'pfstates' => [ @PFSTATE_VARS ]
 );
-@EXPORT_OK = ( @PFSTATE_VARS );
+our @EXPORT_OK = ( @PFSTATE_VARS );
 
 use constant PFSTATE_ERROR       => -1;
 use constant PFSTATE_STARTING    => 0;
@@ -274,7 +271,7 @@ sub main_server_poll {
     # right before select() syscall, but after alarm(), eval scope, etc.
     $self->{child_just_exited} = 0;     
     ($nfound, $timeleft) = select($rout=$rin, undef, $eout=$rin, $tout);
-    $selerr = $!;
+    $selerr = $!  if !defined $nfound || $nfound < 0;
 
   });
 
@@ -299,7 +296,7 @@ sub main_server_poll {
     {
       # this happens if the process is signalled during the select(),
       # for example if someone sends SIGHUP to reload the configuration.
-      # just return inmmediately
+      # just return immediately
       dbg("prefork: select returned err $selerr, probably signalled");
       return;
     }
@@ -515,7 +512,7 @@ sub order_idle_child_to_accept {
   }
   else {
     dbg("prefork: no spare children to accept, waiting for one to complete");
-    return undef;
+    return;
   }
 }
 
@@ -529,15 +526,18 @@ sub wait_for_child_to_accept {
       return 1;     # 1 == success
     }
     if ($state == PFSTATE_ERROR) {
-      return undef;
+      return;
     }
     else {
+      if( Scalar::Util::blessed($self->{server_fh}[0]) eq 'IO::Socket::SSL' ) {
+        warn "prefork: SSL connection protocol error";
+      }
       warn "prefork: ordered child $kid to accept, but they reported state '$state', killing rogue";
       $self->child_error_kill($kid, $sock);
       $self->adapt_num_children();
       sleep 1;
 
-      return undef;
+      return;
     }
   }
 }
@@ -635,7 +635,7 @@ retry_read:
         || (exists &Errno::EWOULDBLOCK && $! == &Errno::EWOULDBLOCK))
     {
       # an error that wasn't non-blocking I/O-related.  that's serious
-      return undef;
+      return;
     }
 
     # ok, we didn't get it first time.  we'll have to start using
@@ -651,18 +651,22 @@ retry_read:
     }
     elsif ($now > $deadline) {
       # timed out!  report failure
-      warn "prefork: sysread(".$sock->fileno.") failed after $timeout secs";
-      return undef;
+      dbg("prefork: sysread(%d) failed after %.1f secs",
+          $sock->fileno, $timeout);
+      return;
     }
     else {
       $tout = $deadline - $now;     # the remaining timeout
       $tout = 1 if ($tout <= 0);    # ensure it's > 0
     }
 
-    dbg("prefork: sysread(".$sock->fileno.") not ready, wait max $tout secs");
+    dbg("prefork: sysread(%d) not ready, wait max %.1f secs",
+        $sock->fileno, $tout);
     my $rin = '';
     vec($rin, $sock->fileno, 1) = 1;
-    select($rin, undef, undef, $tout);
+    my $nfound = select($rin, undef, undef, $tout);
+    defined $nfound && $nfound >= 0
+      or info("prefork: sysread_with_timeout select error: %s", $!);
     goto retry_read;
 
   }
@@ -703,13 +707,15 @@ retry_write:
     warn "prefork: syswrite(".$sock->fileno.") to $targetname failed on try $try";
     if ($try > $numretries) {
       warn "prefork: giving up";
-      return undef;
+      return;
     }
     else {
       # give it 1 second to recover
       my $rout = '';
       vec($rout, $sock->fileno, 1) = 1;
-      select(undef, $rout, undef, 1);
+      my $nfound = select(undef, $rout, undef, 1);
+      defined $nfound && $nfound >= 0
+        or info("prefork: syswrite_with_retry select error: %s", $!);
     }
   }
 
@@ -729,7 +735,7 @@ retry_write:
         || (exists &Errno::EWOULDBLOCK && $! == &Errno::EWOULDBLOCK))
     {
       # an error that wasn't non-blocking I/O-related.  that's serious
-      return undef;
+      return;
     }
 
     warn "prefork: retrying syswrite(): $!";
