@@ -2,6 +2,8 @@
 # imported into main for ease of use.
 package main;
 
+require v5.14.0;
+
 # use strict;
 # use warnings;
 # use re 'taint';
@@ -12,6 +14,7 @@ use File::Basename;
 use File::Copy;
 use File::Path;
 use File::Spec;
+use File::Temp qw(tempdir);
 
 use Test::Builder ();
 use Test::More    ();
@@ -21,8 +24,11 @@ use POSIX qw(WIFEXITED WIFSIGNALED WIFSTOPPED WEXITSTATUS WTERMSIG WSTOPSIG);
 use vars qw($RUNNING_ON_WINDOWS $SSL_AVAILABLE
             $SKIP_SPAMD_TESTS $SKIP_SPAMC_TESTS $NO_SPAMC_EXE
             $SKIP_SETUID_NOBODY_TESTS $SKIP_DNSBL_TESTS
-            $have_inet4 $have_inet6 $spamdhost $spamdport);
+            $have_inet4 $have_inet6 $spamdhost $spamdport
+            $workdir $siterules $localrules $userrules $userstate
+            $keep_workdir $mainpid $spamd_pidfile);
 
+my $sa_code_dir;
 BEGIN {
   require Exporter;
   use vars qw(@ISA @EXPORT @EXPORT_OK);
@@ -45,24 +51,56 @@ BEGIN {
 
   $have_inet4 = eval {
     require IO::Socket::INET;
-    my $sock = IO::Socket::INET->new(LocalAddr => '0.0.0.0', Proto => 'udp');
+    my $sock = IO::Socket::INET->new(LocalAddr => '127.0.0.1', Proto => 'udp');
     $sock->close or die "error closing inet socket: $!"  if $sock;
     $sock ? 1 : undef;
   };
 
   $have_inet6 = eval {
     require IO::Socket::INET6;
-    my $sock = IO::Socket::INET6->new(LocalAddr => '::', Proto => 'udp');
+    my $sock = IO::Socket::INET6->new(LocalAddr => '::1', Proto => 'udp');
     $sock->close or die "error closing inet6 socket: $!"  if $sock;
     $sock ? 1 : undef;
   };
 
   # Clean PATH so taint doesn't complain
-  $ENV{'PATH'} = '/bin:/usr/bin:/usr/local/bin';
-  # Remove tainted envs, at least ENV used in FreeBSD
-  delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};
+  if (!$RUNNING_ON_WINDOWS) {
+    $ENV{'PATH'} = '/bin:/usr/bin:/usr/local/bin';
+    # Remove tainted envs, at least ENV used in FreeBSD
+    delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};
+  } else {
+    # Windows might need non-system directories in PATH to run a Perl installation
+    # The best we can do is clean out obviously bad stuff such as relative paths or \..\
+    my @pathdirs = split(';', $ENV{'PATH'});
+    $ENV{'PATH'} =
+      join(';', # filter for only dirs that are canonical absolute paths that exist
+        map {
+              my $pathdir = $_;
+              $pathdir =~ s/\\*\z//;
+              my $abspathdir = File::Spec->canonpath(Cwd::realpath($pathdir)) if (-d $pathdir);
+              if (defined $abspathdir) {
+                $abspathdir  =~ /^(.*)\z/s;
+                $abspathdir = $1; # untaint it
+              }
+              ((defined $abspathdir) and (lc $pathdir eq lc $abspathdir))?($abspathdir):()
+            }
+          @pathdirs);
+  }
+  
+  # Fix INC to point to absolute path of built SA
+  if (-e 't/test_dir') { $sa_code_dir = 'blib/lib'; }
+  elsif (-e 'test_dir') { $sa_code_dir = '../blib/lib'; }
+  else { die "FATAL: not in or below test directory?\n"; }
+  File::Spec->rel2abs($sa_code_dir) =~ /^(.*)\z/s;
+  $sa_code_dir = $1;
+  if (not -d $sa_code_dir) {
+    die "FATAL: not in expected directory relative to built code tree?\n";
+  }
 }
 
+# use is run at compile time, but after the variable has been computed in the BEGIN block
+use lib $sa_code_dir;
+
 # Set up for testing. Exports (as global vars):
 # out: $home: $HOME env variable
 # out: $cwd: here
@@ -72,6 +110,7 @@ BEGIN {
 #
 sub sa_t_init {
   my $tname = shift;
+  $mainpid = $$;
 
   if ($config{PERL_PATH}) {
     $perl_path = $config{PERL_PATH};
@@ -100,6 +139,31 @@ sub sa_t_init {
   $perl_cmd .= " -T" if !defined($ENV{'TEST_PERL_TAINT'}) or $ENV{'TEST_PERL_TAINT'} ne 'no';
   $perl_cmd .= " -w" if !defined($ENV{'TEST_PERL_WARN'})  or $ENV{'TEST_PERL_WARN'}  ne 'no';
 
+  # Copy directories in PERL5LIB into -I options in perl_cmd because -T suppresses use of PERL5LIB in call to ./spamassassin
+  # If PERL5LIB is empty copy @INC instead because on some platforms like FreeBSD MakeMaker clears PER5LIB and sets @INC
+  # Filter out relative paths, and canonicalize so no symlinks or /../ will be left in untainted result as a nod to security
+  # Since this is only used to run tests, the security considerations are not as strict as with more general situations.
+  my @pathdirs = @INC;
+  if ($ENV{'PERL5LIB'}) {
+    @pathdirs = split($Config{path_sep}, $ENV{'PERL5LIB'});
+  }
+  my $inc_opts =
+    join(' -I', # filter for only dirs that are absolute paths that exist, then canonicalize them
+      map {
+            my $pathdir = $_;
+            my $canonpathdir = File::Spec->canonpath(Cwd::realpath($pathdir)) if ((-d $pathdir) and File::Spec->file_name_is_absolute($pathdir));
+            if (defined $canonpathdir) {
+               $canonpathdir =~ /^(.*)\z/s;
+               $canonpathdir = $1; # untaint it
+            }
+            ((defined $canonpathdir))?($canonpathdir):()
+          }
+         @pathdirs);
+  $perl_cmd .= " -I$inc_opts" if ($inc_opts);
+  
+  # To work in Windows, the perl scripts have to be launched by $perl_cmd and
+  # the ones that are exe files have to be directly called in the command lines
+  
   $scr = $ENV{'SPAMASSASSIN_SCRIPT'};
   $scr ||= "$perl_cmd ../spamassassin.raw";
 
@@ -113,10 +177,10 @@ sub sa_t_init {
   $salearn ||= "$perl_cmd ../sa-learn.raw";
 
   $saawl = $ENV{'SAAWL_SCRIPT'};
-  $saawl ||= "../sa-awl";
+  $saawl ||= "$perl_cmd ../sa-awl";
 
   $sacheckspamd = $ENV{'SACHECKSPAMD_SCRIPT'};
-  $sacheckspamd ||= "../sa-check_spamd";
+  $sacheckspamd ||= "$perl_cmd ../sa-check_spamd";
 
   $spamdlocalhost = $ENV{'SPAMD_LOCALHOST'};
   if (!$spamdlocalhost) {
@@ -124,36 +188,74 @@ sub sa_t_init {
   }
   $spamdhost = $ENV{'SPAMD_HOST'};
   $spamdhost ||= $spamdlocalhost;
-  $spamdport = $ENV{'SPAMD_PORT'};
-  $spamdport ||= probably_unused_spamd_port();
 
   # optimisation -- don't setup spamd test parameters unless we're
   # not skipping all spamd tests and this particular test is called
   # called "spamd_something" or "spamc_foo"
   # We still run spamc tests when there is an external SPAMD_HOST, but don't have to set up the spamd parameters for it
-  if ($SKIP_SPAMD_TESTS or ($tname !~ /spam[cd]/)) {
-    $NO_SPAMD_REQUIRED = 1;
+  if ($tname !~ /spam[cd]/) {
+    $TEST_DOES_NOT_RUN_SPAMC_OR_D = 1;
+  } else {
+    $spamdport = $ENV{'SPAMD_PORT'};
+    $spamdport ||= probably_unused_spamd_port();
   }
 
-  $spamd_cf_args = "-C log/test_rules_copy";
-  $spamd_localrules_args = " --siteconfigpath log/localrules.tmp";
-  $scr_localrules_args =   " --siteconfigpath log/localrules.tmp";
-  $salearn_localrules_args =   " --siteconfigpath log/localrules.tmp";
+  (-f "t/test_dir") && chdir("t");        # run from ..
+  -f "test_dir"  or die "FATAL: not in test directory?\n";
 
-  $scr_cf_args = "-C log/test_rules_copy";
-  $scr_pref_args = "-p log/test_default.cf";
-  $salearn_cf_args = "-C log/test_rules_copy";
-  $salearn_pref_args = "-p log/test_default.cf";
+  unless (-d "log") {
+    mkdir ("log", 0755) or die ("Error creating log dir: $!");
+  }
+  chmod (0755, "log"); # set in case log already exists with wrong permissions
+
+  if (!$RUNNING_ON_WINDOWS) {
+    untaint_system("chacl -B log 2>/dev/null || setfacl -b log 2>/dev/null"); # remove acls that confuse test
+  }
+
+  # clean old workdir if sa_t_init called multiple times
+  if (defined $workdir) {
+    if (!$keep_workdir) {
+      rmtree($workdir);
+    }
+  }
+
+  # individual work directory to make parallel tests possible
+  $workdir = tempdir("$tname.XXXXXX", DIR => "log");
+  die "FATAL: failed to create workdir: $!" unless -d $workdir;
+  $keep_workdir = 0;
+  # $siterules contains all stock *.pre files
+  $siterules = "$workdir/siterules";
+  # $localrules contains all stock *.cf files
+  $localrules = "$workdir/localrules";
+  # $userrules contains user rules
+  $userrules = "$workdir/user.cf";
+  # user_state directory
+  $userstate = "$workdir/user_state";
+
+  mkdir($siterules) or die "FATAL: failed to create $siterules\n";
+  mkdir($localrules) or die "FATAL: failed to create $localrules\n";
+  open(OUT, ">$userrules") or die "FATAL: failed to create $userrules\n";
+  close(OUT);
+  mkdir($userstate) or die "FATAL: failed to create $userstate\n";
+
+  $spamd_pidfile = "$workdir/spamd.pid";
+  $spamd_cf_args = "-C $localrules";
+  $spamd_localrules_args = " --siteconfigpath $siterules";
+  $scr_localrules_args =   " --siteconfigpath $siterules";
+  $salearn_localrules_args =   " --siteconfigpath $siterules";
+
+  $scr_cf_args = "-C $localrules";
+  $scr_pref_args = "-p $userrules";
+  $salearn_cf_args = "-C $localrules";
+  $salearn_pref_args = "-p $userrules";
   $scr_test_args = "";
   $salearn_test_args = "";
-  $set_test_prefs = 0;
+  $set_user_prefs = 0;
   $default_cf_lines = "
-    bayes_path ./log/user_state/bayes
-    auto_whitelist_path ./log/user_state/auto-whitelist
+    bayes_path ./$userstate/bayes
+    auto_welcomelist_path ./$userstate/auto-welcomelist
   ";
 
-  (-f "t/test_dir") && chdir("t");        # run from ..
-
   read_config();
 
   # if running as root, ensure "nobody" can write to it too
@@ -168,77 +270,45 @@ sub sa_t_init {
     $tmp_dir_mode = 0755;
   }
 
-  if (!$NO_SPAMD_REQUIRED) {
-    $NO_SPAMC_EXE = ($RUNNING_ON_WINDOWS &&
+  $NO_SPAMC_EXE = $TEST_DOES_NOT_RUN_SPAMC_OR_D ||
+                  ($RUNNING_ON_WINDOWS &&
                    !$ENV{'SPAMC_SCRIPT'} &&
                    !(-e "../spamc/spamc.exe"));
-    $SKIP_SPAMC_TESTS = ($NO_SPAMC_EXE ||
-                       ($RUNNING_ON_WINDOWS && !$ENV{'SPAMD_HOST'})); 
-    $SSL_AVAILABLE = ((!$SKIP_SPAMC_TESTS) &&  # no SSL test if no spamc
-                    (!$SKIP_SPAMD_TESTS) &&  # or if no local spamd
-                    (untaint_cmd("$spamc -V") =~ /with SSL support/) &&
-                    (untaint_cmd("$spamd --version") =~ /with SSL support/));
-  }
-  # do not remove prior test results!
-  # rmtree ("log");
+  $SKIP_SPAMC_TESTS = ($NO_SPAMC_EXE ||
+                     ($RUNNING_ON_WINDOWS && !$ENV{'SPAMD_HOST'})); 
+  $SSL_AVAILABLE = (!$TEST_DOES_NOT_RUN_SPAMC_OR_D) &&
+                  (!$SKIP_SPAMC_TESTS) &&  # no SSL test if no spamc
+                  (!$SKIP_SPAMD_TESTS) &&  # or if no local spamd
+                  (untaint_cmd("$spamc -V") =~ /with SSL support/) &&
+                  (untaint_cmd("$spamd --version") =~ /with SSL support/);
 
-  unless (-d "log") {
-    mkdir ("log", 0755) or die ("Error creating log dir: $!");
-  }
-  chmod (0755, "log"); # set in case log already exists with wrong permissions
-
-  if (!$RUNNING_ON_WINDOWS) {
-    untaint_system("chacl -B log 2>/dev/null || setfacl -b log 2>/dev/null"); # remove acls that confuse test
-  }
-
-  rmtree ("log/user_state");
-  rmtree ("log/outputdir.tmp");
-
-  rmtree ("log/test_rules_copy");
-  mkdir ("log/test_rules_copy", 0755);
-
-  for $tainted (<../rules/*.cf>, <../rules/*.pm>, <../rules/*.pre>) {
+  for $tainted (<../rules/*.pm>, <../rules/*.pre>, <../rules/languages>) {
     $tainted =~ /(.*)/;
     my $file = $1;
     $base = basename $file;
-    copy ($file, "log/test_rules_copy/$base")
-      or warn "cannot copy $file to log/test_rules_copy/$base: $!";
+    copy ($file, "$siterules/$base")
+      or warn "cannot copy $file to $siterules/$base: $!";
   }
 
-  copy ("data/01_test_rules.pre", "log/test_rules_copy/01_test_rules.pre")
-    or warn "cannot copy data/01_test_rules.cf to log/test_rules_copy/01_test_rules.pre: $!";
-  copy ("data/01_test_rules.cf", "log/test_rules_copy/01_test_rules.cf")
-    or warn "cannot copy data/01_test_rules.cf to log/test_rules_copy/01_test_rules.cf: $!";
-
-  rmtree ("log/localrules.tmp");
-  mkdir ("log/localrules.tmp", 0755);
-
-  for $tainted (<../rules/*.pm>, <../rules/*.pre>) {
+  for $tainted (<../rules/*.cf>) {
     $tainted =~ /(.*)/;
     my $file = $1;
     $base = basename $file;
-    copy ($file, "log/localrules.tmp/$base")
-      or warn "cannot copy $file to log/localrules.tmp/$base: $!";
+    copy ($file, "$localrules/$base")
+      or warn "cannot copy $file to $localrules/$base: $!";
   }
 
-  copy ("../rules/user_prefs.template", "log/test_rules_copy/99_test_default.cf")
-    or die "user prefs copy failed: $!";
+  copy ("data/01_test_rules.pre", "$localrules/01_test_rules.pre")
+    or warn "cannot copy data/01_test_rules.cf to $localrules/01_test_rules.pre: $!";
+  copy ("data/01_test_rules.cf", "$localrules/01_test_rules.cf")
+    or warn "cannot copy data/01_test_rules.cf to $localrules/01_test_rules.cf: $!";
 
-  open (PREFS, ">>log/test_rules_copy/99_test_default.cf")
-    or die "cannot append to log/test_rules_copy/99_test_default.cf: $!";
+  open (PREFS, ">>$localrules/99_test_default.cf")
+    or die "cannot append to $localrules/99_test_default.cf: $!";
   print PREFS $default_cf_lines
-    or die "error writing to log/test_rules_copy/99_test_default.cf: $!";
+    or die "error writing to $localrules/99_test_default.cf: $!";
   close PREFS
-    or die "error closing log/test_rules_copy/99_test_default.cf: $!";
-
-  # create an empty .prefs file
-  open (PREFS, ">>log/test_default.cf")
-    or die "cannot append to log/test_default.cf: $!";
-  close PREFS
-    or die "error closing log/test_default.cf: $!";
-
-  mkdir("log/user_state",$tmp_dir_mode);
-  chmod ($tmp_dir_mode, "log/user_state");  # unaffected by umask
+    or die "error closing $localrules/99_test_default.cf: $!";
 
   $home = $ENV{'HOME'};
   $home ||= $ENV{'WINDIR'} if (defined $ENV{'WINDIR'});
@@ -250,10 +320,26 @@ sub sa_t_init {
   $spamd_run_as_user = ($RUNNING_ON_WINDOWS || ($> == 0)) ? "nobody" : (getpwuid($>))[0] ;
 }
 
-# a port number between 32768 and 65535; used to allow multiple test
+# remove all rules - $localrules/*.cf
+# when you want to only use rules declared inside a specific *.t
+sub clear_localrules {
+  for $tainted (<$localrules/*.cf>) {
+    $tainted =~ /(.*)/;
+    my $file = $1;
+    # Keep some useful, should not contain any rules
+    next if $file =~ /10_default_prefs.cf$/;
+    next if $file =~ /20_aux_tlds.cf$/;
+    # Keep our own tstprefs() or tstlocalrules()
+    next if $file =~ /99_test_prefs.cf$/;
+    next if $file =~ /99_test_rules.cf$/;
+    unlink $file;
+  }
+}
+
+# a port number between 40000 and 65520; used to allow multiple test
 # suite runs on the same machine simultaneously
 sub probably_unused_spamd_port {
-  return 0 if $NO_SPAMD_REQUIRED;
+  return 0 if $SKIP_SPAMD_TESTS;
 
   my $port;
   my @nstat;
@@ -263,11 +349,9 @@ sub probably_unused_spamd_port {
     @nstat = grep(/^\s*tcp/i, <NSTAT>);
     close(NSTAT);
   }
-  my $delta = ($$ % 32768) || int(rand(32768));
-  for (1..10) {
-    $port = 32768 + $delta;
+  for (1..20) {
+    $port = 40000 + int(rand(65500-40000));
     last unless (getservbyport($port, "tcp") || grep(/[:.]$port\s/, @nstat));
-    $delta = int(rand(32768));
   }
   return $port;
 }
@@ -291,31 +375,35 @@ sub sa_t_finish {
 
 sub tstfile {
   my $file = shift;
-  open (OUT, ">log/mail.txt") or die;
+  open (OUT, ">$workdir/mail.txt") or die;
   print OUT $file; close OUT;
 }
 
-sub tstlocalrules {
-  my $lines = shift;
-
-  $set_local_rules = 1;
-
-  open (OUT, ">log/localrules.tmp/00test.cf") or die;
-  print OUT $lines; close OUT;
-}
-
 sub tstprefs {
   my $lines = shift;
 
-  $set_test_prefs = 1;
+  open (OUT, ">$localrules/99_test_prefs.cf") or die;
+  print OUT $lines; close OUT;
+}
+
+sub tstlocalrules {
+  my $lines = shift;
+
+  open (OUT, ">$localrules/99_test_rules.cf") or die;
+  print OUT $lines; close OUT;
+}
+
+sub tstuserprefs {
+  my $lines = shift;
+
+  $set_user_prefs = 1;
 
   # TODO: should we use -p, or modify the test_rules_copy/99_test_default.cf?
   # for now, I'm taking the -p route, since we have to be able to test
   # the operation of user-prefs in general, itself.
 
-  open (OUT, ">log/tst.cf") or die;
+  open (OUT, ">$userrules") or die;
   print OUT $lines; close OUT;
-  $scr_pref_args = "-p log/tst.cf";
 }
 
 # creates a .pre file in the localrules dir to be parsed alongside init.pre
@@ -324,10 +412,25 @@ sub tstprefs {
 sub tstpre {
   my $lines = shift;
 
-  open (OUT, ">log/localrules.tmp/zz_tst.pre") or die;
+  open (OUT, ">$siterules/zz_test.pre") or die;
   print OUT $lines; close OUT;
 }
 
+# remove default compatibility option
+sub disable_compat {
+  my $compat = shift;
+  return unless defined $compat;
+  open (IN, "$siterules/init.pre") or die;
+  open (OUT, ">$siterules/init.pre.new") or die;
+  while (<IN>) {
+    next if $_ =~ /^\s*enable_compat\s+\Q$compat\E(?:\s|$)/i;
+    print OUT $_;
+  }
+  close OUT or die;
+  close IN or die;
+  rename("$siterules/init.pre.new", "$siterules/init.pre");
+}
+
 # Run spamassassin. Calls back with the output.
 # in $args: arguments to run with
 # in $read_sub: callback for the output (should read from <IN>).
@@ -355,14 +458,15 @@ sub sarun {
   my $scrargs = "$scr $args";
   $scrargs =~ s!/!\\!g if ($^O =~ /^MS(DOS|Win)/i);
   print ("\t$scrargs\n");
-  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  (-d "$workdir/d.$testname") or mkdir ("$workdir/d.$testname", 0755);
   
   my $test_number = test_number();
-
-  untaint_system("$scrargs > log/d.$testname/$test_number $post_redir");
+  $current_checkfile = "$workdir/d.$testname/$test_number";
+#print STDERR "RUN: $scrargs\n";
+  untaint_system("$scrargs > $workdir/d.$testname/$test_number $post_redir");
   $sa_exitcode = ($?>>8);
   if ($sa_exitcode != 0) { return undef; }
-  &checkfile ("d.$testname/$test_number", $read_sub) if (defined $read_sub);
+  &checkfile ("$workdir/d.$testname/$test_number", $read_sub) if (defined $read_sub);
   1;
 }
 
@@ -392,14 +496,15 @@ sub salearnrun {
   my $salearnargs = "$salearn $args";
   $salearnargs =~ s!/!\\!g if ($^O =~ /^MS(DOS|Win)/i);
   print ("\t$salearnargs\n");
-  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  (-d "$workdir/d.$testname") or mkdir ("$workdir/d.$testname", 0755);
 
   my $test_number = test_number();
+  $current_checkfile = "$workdir/d.$testname/$test_number";
 
-  untaint_system("$salearnargs > log/d.$testname/$test_number");
+  untaint_system("$salearnargs > $workdir/d.$testname/$test_number");
   $salearn_exitcode = ($?>>8);
   if ($salearn_exitcode != 0) { return undef; }
-  &checkfile ("d.$testname/$test_number", $read_sub) if (defined $read_sub);
+  &checkfile ("$workdir/d.$testname/$test_number", $read_sub) if (defined $read_sub);
   1;
 }
 
@@ -451,14 +556,14 @@ sub spamcrun {
   $spamcargs =~ s!/!\\!g if ($^O =~ /^MS(DOS|Win)/i);
 
   print ("\t$spamcargs\n");
-  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  (-d "$workdir/d.$testname") or mkdir ("$workdir/d.$testname", 0755);
 
   my $test_number = test_number();
 
   if ($capture_stderr) {
-    untaint_system ("$spamcargs > log/d.$testname/out.$test_number 2>&1");
+    untaint_system ("$spamcargs > $workdir/d.$testname/out.$test_number 2>&1");
   } else {
-    untaint_system ("$spamcargs > log/d.$testname/out.$test_number");
+    untaint_system ("$spamcargs > $workdir/d.$testname/out.$test_number");
   }
 
   $sa_exitcode = ($?>>8);
@@ -468,7 +573,7 @@ sub spamcrun {
 
   %found = ();
   %found_anti = ();
-  &checkfile ("d.$testname/out.$test_number", $read_sub) if (defined $read_sub);
+  &checkfile ("$workdir/d.$testname/out.$test_number", $read_sub) if (defined $read_sub);
 
   if ($expect_failure) {
     ($sa_exitcode != 0);
@@ -497,10 +602,10 @@ sub spamcrun_background {
   $spamcargs =~ s!/!\\!g if ($^O =~ /^MS(DOS|Win)/i);
 
   print ("\t$spamcargs &\n");
-  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  (-d "$workdir/d.$testname") or mkdir ("$workdir/d.$testname", 0755);
   
   my $test_number = test_number();
-  untaint_system ("$spamcargs > log/d.$testname/bg.$test_number &") and return 0;
+  untaint_system ("$spamcargs > $workdir/d.$testname/bg.$test_number &") and return 0;
 
   1;
 }
@@ -518,15 +623,15 @@ sub sdrun {
 }
 
 sub recreate_outputdir_tmp {
-  rmtree ("log/outputdir.tmp"); # some tests use this
-  mkdir ("log/outputdir.tmp", $tmp_dir_mode);
-  chmod ($tmp_dir_mode, "log/outputdir.tmp");  # unaffected by umask
+  rmtree ("$workdir/outputdir.tmp"); # some tests use this
+  mkdir ("$workdir/outputdir.tmp", $tmp_dir_mode);
+  chmod ($tmp_dir_mode, "$workdir/outputdir.tmp");  # unaffected by umask
 }
 
 # out: $spamd_stderr
 sub start_spamd {
   return if $SKIP_SPAMD_TESTS;
-  die "NO_SPAMD_REQUIRED in start_spamd! oops" if $NO_SPAMD_REQUIRED;
+  die "TEST_DOES_NOT_RUN_SPAMC_OR_D; in start_spamd! oops" if $TEST_DOES_NOT_RUN_SPAMC_OR_D;
 
   my $spamd_extra_args = shift;
 
@@ -571,13 +676,12 @@ sub start_spamd {
     warn "oops! SATest.pm: a test prefs file was created, but spamd isn't reading it\n";
   }
 
-  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  (-d "$workdir/d.$testname") or mkdir ("$workdir/d.$testname", 0755);
   
   my $test_number = test_number();
-  my $spamd_stdout = "log/d.$testname/spamd.out.$test_number";
-     $spamd_stderr = "log/d.$testname/spamd.err.$test_number";    #  global
-  my $spamd_stdlog = "log/d.$testname/spamd.log.$test_number";
-  my $spamd_pidfile = "log/spamd.pid";
+  my $spamd_stdout = "$workdir/d.$testname/spamd.out.$test_number";
+     $spamd_stderr = "$workdir/d.$testname/spamd.err.$test_number";    #  global
+  my $spamd_stdlog = "$workdir/d.$testname/spamd.log.$test_number";
   my $spamd_forker = $ENV{'SPAMD_FORKER'}   ?
                        $ENV{'SPAMD_FORKER'} :
                      $RUNNING_ON_WINDOWS    ?
@@ -599,7 +703,7 @@ sub start_spamd {
 
   # DEBUG instrumentation to trace spamd processes. See bug 5731 for history
   # if (-f "/home/jm/capture_spamd_straces") {
-  # $spamd_cmd = "strace -ttt -fo log/d.$testname/spamd.strace.$test_number $spamd_cmd";
+  # $spamd_cmd = "strace -ttt -fo $workdir/d.$testname/spamd.strace.$test_number $spamd_cmd";
   # }
 
   unlink ($spamd_stdout, $spamd_stderr, $spamd_stdlog, $spamd_pidfile);
@@ -616,7 +720,11 @@ sub start_spamd {
   sleep $wait ;
   while ($spamd_pid <= 0) {
     my $spamdlog = '';
-    my $pidstr = untaint_cmd("cat $spamd_pidfile 2>/dev/null");
+    my $pidstr;
+    if (open(PID, $spamd_pidfile)) {
+      $pidstr = <PID>;
+      close PID;
+    }
     if ($pidstr) {
        chomp $pidstr;
        $spamd_pid = $pidstr;
@@ -641,7 +749,7 @@ sub start_spamd {
     }
 
     my $sleep = (int($wait++ / 4) + 1);
-    warn "spam_pid not found: Sleeping $sleep - Retry # $retries\n";
+    warn "spam_pid not found: Sleeping $sleep - Retry # $retries\n" if $retries && $retries < 20;
 
     sleep $sleep if $retries > 0;
 
@@ -658,7 +766,7 @@ sub start_spamd {
 
 sub stop_spamd {
   return 0 if ( defined($spamd_already_killed) || $SKIP_SPAMD_TESTS);
-  die "NO_SPAMD_REQUIRED in stop_spamd! oops" if $NO_SPAMD_REQUIRED;
+  die "TEST_DOES_NOT_RUN_SPAMC_OR_D; in stop_spamd! oops" if $TEST_DOES_NOT_RUN_SPAMC_OR_D;
 
   $spamd_pid ||= 0;
   $spamd_pid = untaint_var($spamd_pid);
@@ -688,10 +796,10 @@ sub create_saobj {
 
   # YUCK, these file/dir names should be some sort of variable, at
   # least we keep their definition in the same file for the moment.
-  my %setup_args = ( rules_filename => 'log/test_rules_copy',
-		     site_rules_filename => 'log/localrules.tmp',
-		     userprefs_filename => 'log/test_default.cf',
-		     userstate_dir => 'log/user_state',
+  my %setup_args = ( rules_filename => $localrules,
+		     site_rules_filename => $siterules,
+		     userprefs_filename => $userrules,
+		     userstate_dir => $userstate,
 		     local_tests_only => 1,
                      # debug => 'all',
 		   );
@@ -701,7 +809,6 @@ sub create_saobj {
     $setup_args{$arg} = $args->{$arg};
   }
 
-  # We'll assume that the test has setup INC correctly
   require Mail::SpamAssassin;
 
   my $sa = Mail::SpamAssassin->new(\%setup_args);
@@ -712,7 +819,6 @@ sub create_saobj {
 sub create_clientobj {
   my $args = shift;
 
-  # We'll assume that the test has setup INC correctly
   require Mail::SpamAssassin::Client;
 
   my $client = Mail::SpamAssassin::Client->new($args);
@@ -727,15 +833,11 @@ sub checkfile {
   my $read_sub = shift;
 
   # print "Checking $filename\n";
-  if (!open (IN, "< log/$filename")) {
-    # could be it already contains the "log/" prefix?
-    if (!open (IN, "< $filename")) {
-      warn "cannot open log/$filename or $filename"; return undef;
-    } else {
-      push @files_checked, "$filename";
-    }
+  if (!open (IN, "< $filename")) {
+    warn "cannot open $filename";
+    return undef;
   } else {
-    push @files_checked, "log/$filename";
+    push @files_checked, "$filename";
   }
   &$read_sub();
   close IN;
@@ -743,53 +845,64 @@ sub checkfile {
 
 # ---------------------------------------------------------------------------
 
-sub pattern_to_re {
-  my $pat = shift;
-
-  if ($pat =~ /^\/(.*)\/$/) {
-    return $1;
-  }
-
-  $pat = quotemeta($pat);
-
-  # make whitespace irrelevant; match any amount as long as the
-  # non-whitespace chars are OK.
-  $pat =~ s/\\\s/\\s\*/gs;
-  $pat;
-}
-
-# ---------------------------------------------------------------------------
-
 sub patterns_run_cb {
-  local ($_);
   my $string = shift;
 
-  if (defined $string) {
-    $_ = $string;
-  } else {
-    $_ = join ('', <IN>);
+  if (!defined $string) {
+    $string = join ('', <IN>);
   }
-  $matched_output = $_;
+  $matched_output = $string;
 
   # create default names == the pattern itself, if not specified
+  my %seen;
   foreach my $pat (keys %patterns) {
     if ($patterns{$pat} eq '') {
       $patterns{$pat} = $pat;
     }
+    if ($seen{$patterns{$pat}}++) {
+      die "ERROR: duplicate pattern name found: '$patterns{$pat}'\n";
+    }
+  }
+  %seen = ();
+  foreach my $pat (keys %anti_patterns) {
+    if ($anti_patterns{$pat} eq '') {
+      $anti_patterns{$pat} = $pat;
+    }
+    if ($seen{$anti_patterns{$pat}}++) {
+      die "ERROR: duplicate anti_pattern name found: '$anti_patterns{$pat}'\n";
+    }
   }
 
   foreach my $pat (sort keys %patterns) {
-    my $safe = pattern_to_re ($pat);
-    # print "JMD $patterns{$pat}\n";
-    if ($_ =~ /${safe}/s) {
-      $found{$patterns{$pat}}++;
+    if (index($pat, '(?^') == 0) { # Detect qr// regex, it's a string now
+      if ($string =~ $pat) {
+        $found{$patterns{$pat}}++;
+      }
+    } else {
+      my $re = $pat;
+      $re =~ s/([^A-Za-z_0-9\s])/\\$1/gs; # quotemeta
+      $re =~ s/\s+/\\s+/gs; # normalize whitespace
+      eval { $re = qr/$re/; 1; };
+      if ($@) { die "ERROR: failed to compile regex: '$re'\n"; }
+      if ($string =~ $re) {
+        $found{$patterns{$pat}}++;
+      }
     }
   }
   foreach my $pat (sort keys %anti_patterns) {
-    my $safe = pattern_to_re ($pat);
-    # print "JMD $patterns{$pat}\n";
-    if ($_ =~ /${safe}/s) {
-      $found_anti{$anti_patterns{$pat}}++;
+    if (index($pat, '(?^') == 0) { # Detect qr// regex, it's a string now
+      if ($string =~ $pat) {
+        $found_anti{$anti_patterns{$pat}}++;
+      }
+    } else {
+      my $re = $pat;
+      $re =~ s/([^A-Za-z_0-9\s])/\\$1/gs; # quotemeta
+      $re =~ s/\s+/\\s+/gs; # normalize whitespace
+      eval { $re = qr/$re/; 1; };
+      if ($@) { die "ERROR: failed to compile regex: '$re'\n"; }
+      if ($string =~ $re) {
+        $found_anti{$anti_patterns{$pat}}++;
+      }
     }
   }
 }
@@ -806,8 +919,10 @@ sub ok_all_patterns {
         ok ($found{$type} == 1) or warn "Found more than once: $type at $file line $line.\n";
       }
     } else {
-      warn "\tNot found: $type = $pat at $file line $line.\n";
+      my $typestr = $type eq $pat ? "" : "$type = ";
+      warn "\tNot found: $typestr$pat at $file line $line.\n";
       if (!$dont_ok) {
+        $keep_workdir = 1;
         ok (0);                     # keep the right # of tests
       }
       $wasfailure++;
@@ -817,7 +932,8 @@ sub ok_all_patterns {
     my $type = $anti_patterns{$pat};
     print "\tChecking for anti-pattern $type at $file line $line.\n";
     if (defined $found_anti{$type}) {
-      warn "\tFound anti-pattern: $type = $pat at $file line $line.\n";
+      my $typestr = $type eq $pat ? "" : "$type = ";
+      warn "\tFound anti-pattern: $typestr$pat at $file line $line.\n";
       if (!$dont_ok) { ok (0); }
       $wasfailure++;
     }
@@ -830,6 +946,7 @@ sub ok_all_patterns {
   if ($wasfailure) {
     warn "Output can be examined in: ".
          join(' ', @files_checked)."\n"  if @files_checked;
+    $keep_workdir = 1;
     return 0;
   } else {
     return 1;
@@ -849,7 +966,8 @@ sub skip_all_patterns {
       if ($skip) {
         warn "\tTest skipped: $skip at $file line $line.\n";
       } else {
-        warn "\tNot found: $type = $pat at $file line $line.\n";
+        my $typestr = $type eq $pat ? "" : "$type = ";
+        warn "\tNot found: $typestr$pat at $file line $line.\n";
       }
       skip ($skip, 0);                     # keep the right # of tests
     }
@@ -858,7 +976,8 @@ sub skip_all_patterns {
     my $type = $anti_patterns{$pat};
     print "\tChecking for anti-pattern $type\n";
     if (defined $found_anti{$type}) {
-      warn "\tFound anti-pattern: $type = $pat at $file line $line.\n";
+      my $typestr = $type eq $pat ? "" : "$type = ";
+      warn "\tFound anti-pattern: $typestr$pat at $file line $line.\n";
       skip ($skip, 0);
     }
     else
@@ -937,40 +1056,10 @@ sub conf_bool {
   return 0;                                 # n or 0
 }
 
-sub mk_safe_tmpdir {
-  return $safe_tmpdir if defined($safe_tmpdir);
-
-  my $dir = File::Spec->tmpdir() || 'log';
-
-  # be a little paranoid, since we're using a public tmp dir and
-  # are exposed to race conditions
-  my $retries = 10;
-  my $tmp;
-  while (1) {
-    $tmp = "$dir/satest.$$.".rand(99999);
-    if (!-d $tmp && mkdir ($tmp, 0755)) {
-      if (-d $tmp && -o $tmp) {     # check we own it
-        lstat($tmp);
-        if (-d _ && -o _) {         # double-check, ignoring symlinks
-          last;                     # we got it safely
-        }
-      }
-    }
-
-    die "cannot get tmp dir, giving up" if ($retries-- < 0);
-
-    warn "failed to create tmp dir '$tmp' safely, retrying...";
-    sleep 1;
-  }
-
-  $safe_tmpdir = $tmp;
-  return $tmp;
-}
-
-sub cleanup_safe_tmpdir {
-  if ($safe_tmpdir) {
-    rmtree($safe_tmpdir) or warn "cannot rmtree $safe_tmpdir";
-  }
+sub mk_socket_tempdir {
+  my $dir = tempdir(CLEANUP => 1);
+  die "FATAL: failed to create socket_tempdir: $!" unless -d $dir;
+  return $dir;
 }
 
 sub wait_for_file_to_change_or_disappear {
@@ -1157,4 +1246,11 @@ sub untaint_cmd {
     }
 }
 
+END {
+  # Cleanup workdir (but not if inside forked process)
+  if (defined $workdir && !$keep_workdir && $$ == $mainpid) {
+    rmtree($workdir);
+  }
+}
+
 1;
